module Conditions where

import HXML
import Regexp
import DeleteEmb
import Exp

verifyConditions::[(String,String,[Tree XMLNode])]->([(String,String)],[([String],String)])->[(String,String,[Tree XMLNode])]
verifyConditions [] _ = []
verifyConditions list ([],[]) = list
verifyConditions list (rExp,nCond) = fList
	where
	nList = verifyRegExp rExp list
	fList = verifyOtherCond nCond nList

verifyRegExp::[(String,String)]->[(String,String,[Tree XMLNode])]->[(String,String,[Tree XMLNode])]
verifyRegExp [] list = list
verifyRegExp (e:es) list = flist
	where
	nlist = verifyREaux e list
	flist = verifyRegExp es nlist

verifyREaux _ [] = []
verifyREaux e (a:as)
	|(var==aVar) && (null ntrees) = verifyREaux e as
	|(var==aVar) = (aVar,n,ntrees):(verifyREaux e as)
	|otherwise = a:(verifyREaux e as)
	where
	(aVar,n,trees) = a
	(var,re) = e
	ntrees = verifyMatch re trees
	-- l=length ntrees

verifyMatch re [] = []
verifyMatch re (t:ts) = case t of
	(Tree (ELNode label []) args)
		|ok -> (Tree (TXNode text) []):(verifyMatch re ts)
		|otherwise -> verifyMatch re ts
		where
		text = concat (map extractText args)
		ok=((wholeMatch(searchS re [] text)) == text)
	(Tree (TXNode label) [])
		|ok -> t:(verifyMatch re ts)
		|otherwise -> verifyMatch re ts
		where
		name = getLabelComponent label 1
		ok=((wholeMatch(searchS re [] name)) == name)

extractText (Tree (TXNode label) []) = label++" "
extractText (Tree (ELNode lab []) []) = ""
extractText(Tree (ELNode lab []) args) = concat (map extractText args)

verifyOtherCond [] list = list
verifyOtherCond (c:cs) list = flist
	where
	nlist = verifyCondAux c list
	flist = verifyOtherCond cs nlist

verifyCondAux _ [] = []
verifyCondAux c list = newAssList++nlist
	where
	(varList,cond)=c
	(assList,nlist) = extractVars varList list
	assListMod = organizeAssoc assList
	newAssListMod = evalCond assListMod cond
	newAssList = reOrganize newAssListMod

extractVars::[String]->[(String,String,[Tree XMLNode])]->([(String,String,[Tree XMLNode])],[(String,String,[Tree XMLNode])])
extractVars [] list = ([],list)
extractVars (v:vs) list = ((ass:assoc),flist)
	where
	(ass,nlist) = extractAux v list
	(assoc,flist) = extractVars vs nlist
	extractAux v [] = error ("Attention: variable "++v++"doesn't exists.")
	extractAux v ((lVar,n,trees):ls)
		|v==lVar = ((lVar,n,trees),ls)
		|otherwise = let (nv,nls) = extractAux v ls in (nv,((lVar,n,trees):nls))

evalCond::[[(String,String,String)]]->String->[[(String,String,String)]]
evalCond [] _ = []
evalCond (a:as) cond
	|(null na) = nas
	|otherwise = na:nas
	where
	na = evalAssoc a cond
	nas = evalCond as cond

evalAssoc::[(String,String,String)]->String->[(String,String,String)]
evalAssoc assoc cond
	|res==True = assoc
	|res==False = []
	where
	exp = buildExpression assoc cond
	res = evalExp exp

buildExpression::[(String,String,String)]->String->String
buildExpression [] cond = cond
buildExpression (a:as) cond = "let "++name++"="++value++" in "++(buildExpression as cond)
	where
	(name,_,value)=a

organizeAssoc::[(String,String,[Tree XMLNode])]->[[(String,String,String)]]
organizeAssoc [] = []
organizeAssoc (l:ls) = buildCouples (buildSingleElem l) (organizeAssoc ls)

reOrganize::[[(String,String,String)]]->[(String,String,[Tree XMLNode])]
reOrganize [] = []
reOrganize (l:ls) = rebuildSingleAssoc l (reOrganize ls)

rebuildSingleAssoc [] [] = []
rebuildSingleAssoc ((n,v,s):ls) [] = (n,v,[(Tree (TXNode s) [])]):(rebuildSingleAssoc ls [])
rebuildSingleAssoc ((n1,v1,s1):ls1) ((n2,v2,t2):ls2) = (n1,v1,((Tree (TXNode s1) []):t2)):(rebuildSingleAssoc ls1 ls2)

buildSingleElem::(String,String,[Tree XMLNode])->[[(String,String,String)]]
buildSingleElem (v,n,[]) = []
buildSingleElem (v,n,(t:ts)) = [(v,n,s)]:(buildSingleElem (v,n,ts))
	where
	s=getLabel t
	getLabel (Tree (TXNode lab) []) = lab
	getLabel (Tree (ELNode lab []) []) = lab

buildCouples::[[(String,String,String)]]->[[(String,String,String)]]->[[(String,String,String)]]
buildCouples l [] = l
buildCouples [] _ = []
buildCouples (l1:ls) l2 = nl1++nls
	where
	nl1 = buildCouplesAux l1 l2
	nls = buildCouples ls l2
	buildCouplesAux _ [] = []
	buildCouplesAux l1 (l2:ls2) = (l1++l2):(buildCouplesAux l1 ls2)