module Embedding where

import HXML
import MyParser
import Compute
import ExtendQuery
import DeleteEmb

infinito::Int
infinito = 1000

-- join_path p r ci cd s mapp: p,r = liste di posting, ci/cd=costo di inserimento/cancellazione, s=risultato, mapp = struttura MPtL
join_path::[Pos]->[Pos]->String->Int->Int->MPtL Int->Int->String->[Pos]
join_path [] _ _ _ _ _ _ _ = []
join_path (p:ps) r preQ ci cd mapp sonNum filt
	|(cd == infinito) && (minR == infinito) = join_path ps restR preQ ci cd mapp sonNum filt
	|(minR - distP - ci) > cd = (copiaPPos:ss) ++ (join_path restP restR preQ ci cd mapp sonNum filt)
	|((minR - distP - ci) <= cd) && ((length trees)<sonNum) = error("Attention: the "++(show sonNum)++"th son of the tag specified couldn't be found")
	|(minR - distP - ci) <= cd = (nuoviPos++ss) ++ (join_path restP restR preQ ci cd mapp sonNum filt)
	where
	(trees, minR, nr, restR) = cicloR r p [] infinito filt []
	(preP, distP, boundP, renP, _, embTreeP) = p
	cmin = min cd (minR - distP - ci)
	name = getMapName preP mapp
	(ss, restP) = find_SelfDesc p ps nr preQ ci cd mapp sonNum filt
	nuoviPos = buildPos preP distP boundP renP cmin name preQ sonNum trees
	copiaPPos = (preP, distP, boundP, renP, cmin, (setLabel embTreeP (preQ++"^"++(show preP))))

buildPos preP distP boundP renP cmin name preQ num trees
	|num == 0 = map (bp preP distP boundP renP cmin name preQ) trees
	|(num < 0) && (lts == 1) = [res]
	|(num < 0) = buildPos preP distP boundP renP cmin name preQ num (tail trees)
	|num == 1 = [res]
	|otherwise = buildPos preP distP boundP renP cmin name preQ (num-1) (tail trees)
	where
	lts = length trees
	res = bp preP distP boundP renP cmin name preQ (head trees)
	bp preP distP boundP renP cmin name preQ tree = (preP, distP, boundP, renP, cmin, (Tree (ELNode (name++"^"++preQ++"^"++(show preP)) []) [tree]))

setLabel (Tree (ELNode label []) args) s = (Tree (ELNode (label++"^"++s) []) args)
setLabel (Tree (TXNode label) []) s = (Tree (TXNode (label++"^"++s)) [])

-- find_SelfDesc p ps nr ci cd s map: funzione che controlla se in "ps" ci sono discendenti di "p" (albero ricorsivo). 
-- Se ci sono chiama join_path sulla porzione di ps da 1 ad "i" che contiene tali discendenti e su 
-- "nr" che  gi la porzione di "r" che contiene i discendenti di "p" in "r".

find_SelfDesc::Pos->[Pos]->[Pos]->String->Int->Int->MPtL Int->Int->String->([Pos],[Pos])
find_SelfDesc _ [] _ _ _ _ _ _ _ = ([],[])
find_SelfDesc p ps [] _ _ _ _ _ _ = ([],ps)
find_SelfDesc p ps nr preQ ci cd map num filt
	|i == 1 = ([],ps)
	|otherwise = ((join_path nps nr preQ ci cd map num filt),rest)
	where
	(i, nps, rest) = scan p ps [] 1

-- scan p ps nps i: funzione che scandisce la lista "ps" alla ricerca di discendenti di "p", nps ed "i" sono il risultato

scan::Pos->[Pos]->[Pos]->Int->(Int,[Pos],[Pos])
scan p [] nps i = (i, (reverse nps), [])
scan p (p2:ps) nps i
	|preP2 <= boundP = scan p ps (p2:nps) (i+1)
	|preP2 > boundP = (i,(reverse nps),(p2:ps))
	where
	(_, _, boundP, _, _, _) = p
	(preP2, _, _, _, _, _) = p2

-- cicloR r p tree cmin nr: funzione che dato "p" un posting, cerca nella lista di posting "r" 
-- i discendenti di "p" e accumula negli altri parametri il risultato. 
-- "tree"  l'embedding tree del discendente a minor costo, "cmin"  tale costo,
-- "nr"  la porzione di "r" che contiene i discendenti di "p".

-- viewR (tree,c,nr) = do {putStrLn ("Tree: "++(show tree)++"\nCost: "++(show c)++"\n");
--			     mymap view nr;}

cicloR::[Pos]->Pos->[Tree XMLNode]->Int->String->[Pos]->([Tree XMLNode],Int,[Pos],[Pos])
cicloR [] _ tree cmin _ nr = (tree, cmin, nr, [])
cicloR (r:rs) pos tree cmin filt nr
	|(preR > preP) && (preR <= boundP) && (cost < cmin) = cicloR rs pos [embTreeR] cost filt (r:nr)
	|(preR > preP) && (preR <= boundP) && (cost == cmin) && (filt=="all") = cicloR rs pos (tree++[embTreeR]) cmin filt (r:nr)
	|(preR > preP) && (preR <= boundP) && (cost == cmin) && (filt=="best") = cicloR rs pos tree cmin filt (r:nr)
	|(preR > preP) && (preR <= boundP) = cicloR rs pos tree cmin filt (r:nr)
	|(preR > preP) = (tree, cmin, nr, (r:rs))
	|preR <= preP = cicloR rs pos tree cmin filt nr
	where
	(preR,distR,_,_,embcostR,embTreeR) = r
	(preP,_,boundP,_,_,_) = pos
	cost = embcostR + distR

-- adjust_attributes tree, funzione che inserisce come attributi di un nodo i figli la cui etichetta inizia per '@'
adjust_attributes (Tree (ELNode label []) args) = (Tree (ELNode label attr) nArgs)
	where
	(attr, nArgs) = aux2 args
adjust_attributes (Tree (TXNode label) []) = (Tree (TXNode label) [])

aux2 [] = ([], [])
aux2 (t:ts) = (attr++attrs, son++sons)
	where
	(attr, son) = aux t
	(attrs, sons) = aux2 ts

aux tree =
	case tree of
		(Tree (ELNode ('@':label) []) [son]) -> ([(label, (extract_text son))], [])
			where
			extract_text (Tree (TXNode lab) []) = lab
       		(Tree (ELNode label []) args) -> ([], [(Tree (ELNode label attr) nArgs)])
		       	where
			(attr,nArgs) = aux2 args
		(Tree (TXNode label) []) -> ([], [(Tree (TXNode label) [])])

add_rencost::[Pos]->[Pos]
add_rencost [] = []
add_rencost (p:ps) = nuovoPos : (add_rencost ps)
	where 
	(preP, distP, boundP, renP, embcostP, embTreeP) = p
	nuovoPos = (preP, distP, boundP, renP, embcostP+renP, embTreeP)

add_delcost::[Pos]->Int->[Pos]
add_delcost [] _ = []
add_delcost (p:ps) c 
	|c==0 = (p:ps)
	|otherwise = nuovoPos : (add_delcost ps c)
	where 
	(preP, distP, boundP, renP, embcostP, embTreeP) = p
	nuovoPos = (preP, distP, boundP, renP, embcostP+c, embTreeP)

evaluateQuery::Tree XMLNode->MPtL String->SumDelCost->[Pos]->String->Index->Index->MPtL Int->Int->MPtL String->String->String->[Pos]
evaluateQuery t qMap sdc [] preQ indexEL indexTX mapp delC sType filter defaultType = case t of
	(Tree (ELNode label _) (arg:_)) 
		|i==0 -> let (lType,nSType)=(searchType label defaultType sType) 
			 in add_rencost (evaluate arg qMap sdc (getPresences indexEL (getMapName label qMap) lType) preQ indexEL indexTX mapp delC nSType filter defaultType)
		|otherwise -> error("Attention: the branch information couldn't be applied to the query root")
		where
		(lab,i) = getInfofromLabel (getMapName label qMap)
	(Tree (ELNode label _) []) 
		|i==0 -> let (lType,_)=(searchType label defaultType sType) 
			 in add_rencost (map (addPre preQ) (getPresences indexEL (getMapName label qMap) lType))
		|otherwise -> error("Attention: the branch information couldn't be applied to the query root")
		where
		(lab,i) = getInfofromLabel (getMapName label qMap)
	(Tree (TXNode label) []) 
		|i==0 -> let (lType,_)=(searchType label defaultType sType) 
			 in add_rencost (map (addPre preQ) (getPresences indexTX (getMapName label qMap) lType))
		|otherwise -> error("Attention: the branch information couldn't be applied to the query root")
		where
		(lab,i) = getInfofromLabel (getMapName label qMap)

evaluate::Tree XMLNode->MPtL String->SumDelCost->[Pos]->String->Index->Index->MPtL Int->Int->MPtL String->String->String->[Pos]
evaluate _ _ _ [] _ _ _ _ _ _ _ _ = []
evaluate t qMap sdc p preQ indexEL indexTX mapp delC sType filt defaultType =
	case t of
		(Tree (ELNode label attr) []) -> let (lType,_)=(searchType label defaultType sType)
						     (lab,i) = getInfofromLabel (getMapName label qMap)
						     sdcost = (getSDC label sdc)
					         in join_path p (map (addPre label) (add_rencost (add_delcost (getPresences indexEL lab lType) sdcost)) ) preQ inscost (delC+sdcost) mapp i filt
		(Tree (ELNode label attr) (arg:args))
			|((getMapName label qMap) == "and") -> sumSons qMap sdc preQ p (getSDC label sdc) (evaluate arg qMap sdc p preQ indexEL indexTX mapp delC sType filt defaultType) args indexEL indexTX mapp delC sType filt defaultType
			|((getMapName label qMap) == "or") -> minSons qMap sdc preQ p (getSDC label sdc) (evaluate arg qMap sdc p preQ indexEL indexTX mapp delC sType filt defaultType) args indexEL indexTX mapp delC sType filt defaultType
			|((getMapName label qMap) == "xor") -> minXorSons qMap sdc preQ p (getSDC label sdc) (evaluate arg qMap sdc p preQ indexEL indexTX mapp delC sType filt defaultType) args indexEL indexTX mapp delC sType filt defaultType
			|otherwise -> let (lType,nSType)=(searchType label defaultType sType)
					  (lab,i) = getInfofromLabel (getMapName label qMap)
				      in join_path p (add_rencost (evaluate arg qMap sdc (getPresences indexEL lab lType) label indexEL indexTX mapp delC nSType filt defaultType)) preQ inscost infinito mapp i filt
		(Tree (TXNode label) []) -> let (lType,_)=(searchType label defaultType sType)
						(lab,i) = getInfofromLabel (getMapName label qMap)
						sdcost = (getSDC label sdc)
					    in join_path p (map (addPre label) (add_rencost (add_delcost (getPresences indexTX lab lType) sdcost)) ) preQ inscost (delC+sdcost) mapp i filt
	where
	sumSons _ _ _ _ _ r [] _ _ _ _ _ _ _ = r
	sumSons qm sdc pq p c r (a:as) ie it m cd t filter df = case filter of 
		"best" -> sumSons qm sdc pq p c (compute_sum_best r (evaluate a qm sdc p pq ie it m cd t filter df) c) as ie it m cd t filter df
		"all" -> sumSons qm sdc pq p c (compute_sum_all r (evaluate a qm sdc p pq ie it m cd t filter df) c) as ie it m cd t filter df
	minSons _ _ _ _ _ r [] _ _ _ _ _ _ _ = r
	minSons qm sdc pq p c r (a:as) ie it m cd t filter df = case filter of
		"best" -> minSons qm sdc pq p c (compute_min_best r (evaluate a qm sdc p pq ie it m cd t filter df) c) as ie it m cd t filter df
		"all" -> minSons qm sdc pq p c (compute_min_all r (evaluate a qm sdc p pq ie it m cd t filter df) c) as ie it m cd t filter df
	minXorSons _ _ _ _ _ r [] _ _ _ _ _ _ _ = r
	minXorSons qm sdc pq p c r (a:as) ie it m cd t filter df = case filter of
		"best" -> minXorSons qm sdc pq p c (compute_xor_best r (evaluate a qm sdc p pq ie it m cd t filter df) c) as ie it m cd t filter df
		"all" -> minXorSons qm sdc pq p c (compute_xor_all r (evaluate a qm sdc p pq ie it m cd t filter df) c) as ie it m cd t filter df
	
addPre::String->Pos->Pos
addPre s r = (p,d,b,rc,c,nt)
	where
	(p,d,b,rc,c,t) = r
	nt = setLabel t (s++"^"++(show p))

inverseType "N" = "E"
inverseType "E" = "N"

searchType::String->String->MPtL String->(String,MPtL String)
searchType _ defaultType []  = (defaultType,[])
searchType label defaultType (x:xs)
	|l1==l2 = ((inverseType defaultType),xs)
	|l1<l2 = (defaultType,(x:xs))
	|l1>l2 = searchType label defaultType xs
	where
	(lab,_)=x
	l1 = (read label) ::Int
	l2 = (read lab) ::Int

getInfofromLabel::String->(String,Int)
getInfofromLabel s
	|j==0 = (t,0)
	|nj==0 = error("Attention: inconsistent branch information found")
	|otherwise = (t,nj)
	where
	(t,j,r) = getLabelName s
	nj = getNumber r

getLabelName [] = ([],0,[])
getLabelName (x:xs)
	|x=='[' = ([],1,(readNum xs))
	|otherwise = ((x:t),j,r)
	where
	(t,j,r) = getLabelName xs
	readNum "]" = []
	readNum (y:ys) = y:(readNum ys)

getNumber::String->Int
getNumber xs
	|xs==[] = error("Attention: inconsistent branch information found")
	|xs=="last" = -1
	|otherwise = (read xs) ::Int

getPresences::Index->Name->String->[Pos]
getPresences index name "E" = getRef index name "E"
getPresences index name "N"
	|(not (null resIndex)) = resIndex
	|(null syn) = []
	|otherwise = resPos
	where
	resIndex = getRef index name "N"
	syn = classOfSyn name
	resPos = getPosOfSyn index (tail syn)

-- versione veloce non per ontologie
getPosOfSyn _ [] = []
getPosOfSyn index (syn:xs)
	|(not (null pos)) = setRenCost 2 pos
	|otherwise = getPosOfSyn index xs
	where
	(n,c) = syn
	pos = getRef index n "N"

setRenCost c [] = []
setRenCost c (p:ps) = (pre,dist,bound,c,emb,tree):(setRenCost c ps)
	where
	(pre,dist,bound,ren,emb,tree) = p
{-
--versione completa per ontologie
getPosOfSyn _ [] = []
getPosOfSyn index (syn:xs)
	|(not (null pos)) = (setRenCost c pos)++(getPosOfSyn index xs)
	|otherwise = getPosOfSyn index xs
	where
	(n,c) = syn
	pos = getRef index n "E"
-}
-- Costi:
inscost::Int
inscost = 2

view (p,_,_,_,c,tree) = putStrLn ("Embedding cost: "++(show c)++"\nEmbedding position: "++(show p)++"\nEmbedding tree:\n"++(show (adjust_attributes tree))++"\n");
{-
es = join_path [(1,0,6,0,0,(Tree (ELNode "a" []) []))] [(2,2,2,0,0,(Tree (TXNode "b") [])),(4,4,4,0,0,(Tree (TXNode "c") [])),(7,4,8,0,0,(Tree (ELNode "d" []) []))] "1" 2 5 [(1,"a")] 0 "all"
es2 = show a
	where
	(a,b,c)=cicloR [(2,2,2,0,0,(Tree (TXNode "b") [])),(4,4,4,0,0,(Tree (TXNode "c") [])),(7,4,8,0,0,(Tree (ELNode "d" []) []))] (1,0,6,0,0,(Tree (ELNode "a" []) [])) [] infinito "all" []
-}